{% if on_saltstack %}
<h4>SaltStack training</h4>
<p>
<a href="http://www.saltstack.com/training/">
Now offering remote attendee training!
{# <img src="{{ pathto('_static/images/salt-training.png', 1) }}" alt="" style="padding-bottom: 1em;"> #}
</a>
</p>
{% endif %}
