<secret ephemeral='no' private='yes'>
    <description>{{ description }}</description>
    {% if type == 'chap' %}
    <usage type='iscsi'>
        <target>{{ usage }}</target>
    </usage>
    {% elif type == 'ceph' %}
    <usage type='ceph'>
        <name>{{ usage }}</name>
    </usage>
    {% endif %}
</secret>
