<volume{% if type %} type='{{ type }}'{% endif %}>
  <name>{{ name }}</name>
  <source>
  </source>
  <capacity unit='KiB'>{{ size }}</capacity>
  <allocation unit='KiB'>{{ allocation }}</allocation>
  <target>
    {%- if format %}
    <format type='{{ format }}'/>
    {%- endif %}
    {%- if target.permissions -%}
    <permissions>
      {%- if target.permissions.get('mode') %}
      <mode>{{ target.permissions.mode }}</mode>
      {%- endif %}
      {%- if target.permissions.get('owner') %}
      <owner>{{ target.permissions.owner }}</owner>
      {%- endif %}
      {%- if target.permissions.get('group') %}
      <group>{{ target.permissions.group }}</group>
      {%- endif %}
      {%- if target.permissions.get('label') %}
      <label>{{ target.permissions.label }}</label>
      {%- endif %}
    </permissions>
    {%- endif %}
    {%- if target.nocow %}
    <nocow/>
    {%- endif %}
  </target>
  {%- if backingStore %}
  <backingStore>
    <path>{{ backingStore.path }}</path>
    {%- if backingStore.format %}
    <format type='{{ backingStore.format }}'/>
    {%- endif %}
  </backingStore>
  {%- endif %}
</volume>
