{%- for route in routes -%}
{% if route.name %}# {{route.name}} {%- endif %}
{{ route.ipaddr }}
{%- if route.gateway %}	{{route.gateway}}{% else %}	-{% endif %}
{%- if route.netmask %}	{{route.netmask}}{% else %}	-{% endif %}
{%- if route.dev %}	{{route.dev}}{% else %}{%- if iface and iface != "routes" %}	{{iface}}{% else %}	-{% endif %}{% endif %}
{%- if route.metric %}	metric {{route.metric}} {%- endif %}
{% endfor -%}
